/******************************************************************************/
/*                                                                            */
/*    The Multilevel Grid File (MLGF)                                         */
/*    Version 4.0                                                             */
/*                                                                            */
/*    Developed by Professor Kyu-Young Whang et al. (1994-2016)               */
/*                                                                            */
/*    Advanced Information Technology Research Center (AITrc)                 */
/*    Korea Advanced Institute of Science and Technology (KAIST)              */
/*                                                                            */
/*    e-mail: kywhang@gmail.com                                               */
/*                                                                            */
/*    Bibliography:                                                           */
/*    [1] Whang, K. and Krishnamurthy, R., "The Multilevel Grid File - A      */
/*        Dynamic Hierarchical Multidimensional File Structure," In Proc. 2nd */
/*        Int'l Conf. on Database Systems for Advanced Applications, pp.      */
/*        449-459, 1991.                                                      */
/*    [2] Whang, K., Kim, S., and Wiederhold, G., "Dynamic Maintenance of     */
/*        Data Distribution for Selectivity Estimation," The VLDB Journal,    */
/*        Vol. 3, No. 1, pp. 29-51, 1994.                                     */
/*    [3] Lee, J., Lee, Y., and Whang, K., "A Region Splitting Strategy for   */
/*        Physical Database Design of Multidimensional File Organizations,"   */
/*        In Proc. 23rd Int'l Conf. on Very Large Data Bases, pp. 416-425,    */
/*        1997.                                                               */
/*    [4] Song, J., Whang, K., Lee, Y., Lee, M., and Kim, S., "Spatial Join   */
/*        Processing Using Corner Transformation, IEEE Transactions on        */
/*        Knowledge and Data Engineering (TKDE), Vol. 11, No. 4, pp. 688-695, */
/*        1999.                                                               */
/*    [5] Song, J. et al., "The Clustering Property of Corner Transformation  */
/*        for Spatial Database Applications," Information and Software        */
/*        Technology, Vol. 44, No. 7, pp. 419-429, 2002.                      */
/*    [6] Lee, M., Whang, K., Han, W., and Song, I., "Transform-Space View:   */
/*        Performing Spatial Join in the Transform Space Using Original-Space */
/*        Indexes," IEEE Transactions on Knowledge and Data Engineering       */
/*        (TKDE), Vol 18, No. 2, pp. 245-260, 2006.                           */
/*    [7] Dai, H., Whang, K., and Su, H., "Locality of Corner Transformation  */
/*        for Multidimensional Spatial Access Methods," Electronic Notes in   */
/*        Theoretical Computer Science, Vol. 212, pp. 133-148, 2008.          */
/*                                                                            */
/******************************************************************************/

/* 
 *  Description : 
 *      Write a page given buffer pageid into the disk. Construct a hash key
 *      using the given SegmentID and page id in order to look up the buffer
 *      page in the buffer pool.  If it is successfully found, open the file
 *      corresponding to SegmentID,  and force it out to the disk using DsM, 
 *      especially DsM_FlushPage().
 *
 *  Return Values : None
 *
 *  Errors : 
 *      eNOERROR : No error
 *      some errors : caused by BfM_LookUp(), DsM_WritePage(), etc.
 */


#include <DsM.h>
#include <BfM.h>


int BfM_FlushPage(int SegmentID, PageID* pageid)
{
    int e;                    /* for errors */
    int index;                /* for an index */
    BfMHashKey key;           /* a hash key */

    /* Construct a hash key using the given parameters */
    key.SegID = SegmentID;
    key.pageid = *pageid;

    /* Look up the key in the hash table and get a corresponding index */
    index = BfM_LookUp(&key);

    if(index < 0)   /* errors in BfM_LookUp() */
        return(index);

    /* write a page into the disk */
    e = DsM_WritePage(SegmentID, pageid, &(BfM_BufferPool[index * BfM_PAGESIZE]));

    if(e < 0)   /* errors in DsM_WritePage() */
        return e;

    /* clear the dirty bit */
    BfM_BufferTable[index].bits ^= DIRTY;

    return(eNOERROR);

}  /* BfM_FlushPage */
