/******************************************************************************/
/*                                                                            */
/*    The Multilevel Grid File (MLGF)                                         */
/*    Version 4.0                                                             */
/*                                                                            */
/*    Developed by Professor Kyu-Young Whang et al. (1994-2016)               */
/*                                                                            */
/*    Advanced Information Technology Research Center (AITrc)                 */
/*    Korea Advanced Institute of Science and Technology (KAIST)              */
/*                                                                            */
/*    e-mail: kywhang@gmail.com                                               */
/*                                                                            */
/*    Bibliography:                                                           */
/*    [1] Whang, K. and Krishnamurthy, R., "The Multilevel Grid File - A      */
/*        Dynamic Hierarchical Multidimensional File Structure," In Proc. 2nd */
/*        Int'l Conf. on Database Systems for Advanced Applications, pp.      */
/*        449-459, 1991.                                                      */
/*    [2] Whang, K., Kim, S., and Wiederhold, G., "Dynamic Maintenance of     */
/*        Data Distribution for Selectivity Estimation," The VLDB Journal,    */
/*        Vol. 3, No. 1, pp. 29-51, 1994.                                     */
/*    [3] Lee, J., Lee, Y., and Whang, K., "A Region Splitting Strategy for   */
/*        Physical Database Design of Multidimensional File Organizations,"   */
/*        In Proc. 23rd Int'l Conf. on Very Large Data Bases, pp. 416-425,    */
/*        1997.                                                               */
/*    [4] Song, J., Whang, K., Lee, Y., Lee, M., and Kim, S., "Spatial Join   */
/*        Processing Using Corner Transformation, IEEE Transactions on        */
/*        Knowledge and Data Engineering (TKDE), Vol. 11, No. 4, pp. 688-695, */
/*        1999.                                                               */
/*    [5] Song, J. et al., "The Clustering Property of Corner Transformation  */
/*        for Spatial Database Applications," Information and Software        */
/*        Technology, Vol. 44, No. 7, pp. 419-429, 2002.                      */
/*    [6] Lee, M., Whang, K., Han, W., and Song, I., "Transform-Space View:   */
/*        Performing Spatial Join in the Transform Space Using Original-Space */
/*        Indexes," IEEE Transactions on Knowledge and Data Engineering       */
/*        (TKDE), Vol 18, No. 2, pp. 245-260, 2006.                           */
/*    [7] Dai, H., Whang, K., and Su, H., "Locality of Corner Transformation  */
/*        for Multidimensional Spatial Access Methods," Electronic Notes in   */
/*        Theoretical Computer Science, Vol. 212, pp. 133-148, 2008.          */
/*                                                                            */
/******************************************************************************/

/*
 * Module: MLGF_CloseIndex.c
 *
 * Description:
 *  Close the MLGF index.
 *
 * Exports:
 *  Four MLGF_CloseIndex(Two)
 *
 * Returns:
 *  Error code
 *    eBADPARAMETER
 *    eSYSERR
 */


#include "common.h"


Four MLGF_CloseIndex(
	Two		mlgfd)			/* IN mlgf file descriptor */
{
	FreePageList	*freePagePtr;	/* a pointer to free page list */
	FreePageList	*tmp;			/* a pointer to free page list */
	Four			count;			/* variable to count the number of free pages */


	/* check the parameters */
	if(!IS_VALID_MLGFD(mlgfd)) ERR(eBADPARAMETER);

	if((-- OPENFILE_COUNT(mlgfd)) > 0) return(eNOERROR);

#ifdef MBR_MLGF_BUFFER
	/* flush buffer before writing */
	BfM_Flush();	
#endif	/* MBR_MLGF_BUFFER */

	/* go to end of file */
	if(lseek(OPENFILE_OS_FD(mlgfd), 0L, L_END) <0) {
		perror(" lseek (in MLGFclose)");
		ERR(eSYSERR);
	}

	/* write control infomation */
	if(write(OPENFILE_OS_FD(mlgfd), (char *)&OPENFILE_CONTROL(mlgfd),
		    sizeof(Control)) != sizeof(Control)) {
		perror(" write (in MLGFclose)");
		ERR(eSYSERR);
	}

	/* write pageCount */
	if(write(OPENFILE_OS_FD(mlgfd), (char *)&OPENFILE_PAGECOUNT(mlgfd),
		    sizeof(PageID)) != sizeof(PageID)) {
		perror(" write  (in MLGFclose)");
		ERR(eSYSERR);
	}

	/* let freePagePtr point the start of free page list */
	freePagePtr = OPENFILE_FREEPAGELIST(mlgfd).next;

	/* initialize count */
	count = 0;

	/* write free page information */
	while(freePagePtr != NULL) {
		/* write free page ID */
		if(write(OPENFILE_OS_FD(mlgfd),
			    (char *)&(freePagePtr->pid),
			    sizeof(PageID)) != sizeof(PageID)) {
			perror(" write (in MLGFclose)");
			ERR(eSYSERR);
		}

		/* see next freePageID */
		tmp = freePagePtr;
		freePagePtr = freePagePtr->next;

		/* free memory */
		free(tmp);

		/* count the number of free pages */
		count ++;
	}
		
	/* save the number of free pages */
	if(write(OPENFILE_OS_FD(mlgfd), (char *)&count,
		    sizeof(Four)) != sizeof(Four)) {
		perror(" write  (in MLGFclose)");
		ERR(eSYSERR);
	}

#ifdef MBR_MLGF_BUFFER
    /* flush buffer */
    BfM_Flush();	
    /* close file */
    DsM_CloseSegment(OPENFILE_DSMSEGID(mlgfd));
#else
	/* close file */
	close( OPENFILE_OS_FD(mlgfd));
#endif	/* MBR_MLGF_BUFFER */
	
	OPENFILE_INO(mlgfd) = NIL;

	return(eNOERROR);
} /* MLGF_CloseIndex() */
