/******************************************************************************/
/*                                                                            */
/*    The Multilevel Grid File (MLGF)                                         */
/*    Version 4.0                                                             */
/*                                                                            */
/*    Developed by Professor Kyu-Young Whang et al. (1994-2016)               */
/*                                                                            */
/*    Advanced Information Technology Research Center (AITrc)                 */
/*    Korea Advanced Institute of Science and Technology (KAIST)              */
/*                                                                            */
/*    e-mail: kywhang@gmail.com                                               */
/*                                                                            */
/*    Bibliography:                                                           */
/*    [1] Whang, K. and Krishnamurthy, R., "The Multilevel Grid File - A      */
/*        Dynamic Hierarchical Multidimensional File Structure," In Proc. 2nd */
/*        Int'l Conf. on Database Systems for Advanced Applications, pp.      */
/*        449-459, 1991.                                                      */
/*    [2] Whang, K., Kim, S., and Wiederhold, G., "Dynamic Maintenance of     */
/*        Data Distribution for Selectivity Estimation," The VLDB Journal,    */
/*        Vol. 3, No. 1, pp. 29-51, 1994.                                     */
/*    [3] Lee, J., Lee, Y., and Whang, K., "A Region Splitting Strategy for   */
/*        Physical Database Design of Multidimensional File Organizations,"   */
/*        In Proc. 23rd Int'l Conf. on Very Large Data Bases, pp. 416-425,    */
/*        1997.                                                               */
/*    [4] Song, J., Whang, K., Lee, Y., Lee, M., and Kim, S., "Spatial Join   */
/*        Processing Using Corner Transformation, IEEE Transactions on        */
/*        Knowledge and Data Engineering (TKDE), Vol. 11, No. 4, pp. 688-695, */
/*        1999.                                                               */
/*    [5] Song, J. et al., "The Clustering Property of Corner Transformation  */
/*        for Spatial Database Applications," Information and Software        */
/*        Technology, Vol. 44, No. 7, pp. 419-429, 2002.                      */
/*    [6] Lee, M., Whang, K., Han, W., and Song, I., "Transform-Space View:   */
/*        Performing Spatial Join in the Transform Space Using Original-Space */
/*        Indexes," IEEE Transactions on Knowledge and Data Engineering       */
/*        (TKDE), Vol 18, No. 2, pp. 245-260, 2006.                           */
/*    [7] Dai, H., Whang, K., and Su, H., "Locality of Corner Transformation  */
/*        for Multidimensional Spatial Access Methods," Electronic Notes in   */
/*        Theoretical Computer Science, Vol. 212, pp. 133-148, 2008.          */
/*                                                                            */
/******************************************************************************/

/*
 * Module: MLGF_Dump.c
 *
 * Description:
 *  Dump the whole structure of mlgf file.
 *  If recordOnly is TRUE, dump all the records only in the bulk export format.
 *
 * Exports:
 *  Four MLGF_Dump(Two, Boolean, FILE*)
 *
 * Returns:
 *  Error code
 *    eBADPARAMETER
 */


#include	"common.h"


int		nDirPages;
int		nLeafPages;
float	dirPageOccupancyRatio;
float	leafPageOccupancyRatio;


/* Internal Fucntion Prototypes */
Four mlgf_DumpRecursive(Two, PageID, Boolean, FILE*);


Four MLGF_Dump(
	Two			mlgfd,			/* IN mlgf file descriptor */
	Boolean		recordOnly,		/* IN whether to dump the record only in the bulk export format */
	FILE		*fd)			/* OUT output file descriptor */
{
	Four	e;

	/* check parameters */
	if ( !IS_VALID_MLGFD(mlgfd))
		return(eBADPARAMETER);
	
	if (!recordOnly) {
		nDirPages = 0;
		nLeafPages = 0;
		dirPageOccupancyRatio = 0;
		leafPageOccupancyRatio = 0;
	}
    
	e = mlgf_DumpRecursive( mlgfd, OPENFILE_CONTROL(mlgfd).rootPage, recordOnly, fd);
	if (e < 0) ERR(e);

	if (!recordOnly) {
		dirPageOccupancyRatio = dirPageOccupancyRatio / nDirPages;
		leafPageOccupancyRatio = leafPageOccupancyRatio / nLeafPages;
		printf("\n\n\nnDirPages = %d (dirPageOccupancyRatio = %f)\nnLeafPages = %d (leafPageOccupancyRatio = %f)\n",	nDirPages, dirPageOccupancyRatio, nLeafPages, leafPageOccupancyRatio);
	}

	return(eNOERROR);
}



Four mlgf_DumpRecursive(
	Two		mlgfd, 			/* IN mlgf file descriptor */
	PageID	pageID,			/* IN page ID */
	Boolean	recordOnly,		/* IN whether to dump the record only in the bulk export format */
	FILE	*fd)			/* OUT output file descriptor */
{
	mlgf_DirectoryEntry		*entryPtr;		/* pointer to entry */
	PageID					nextPageID;		/* next page ID */
	Two						i;				/* index */
	Four					e;				/* error code */
	mlgf_Page	*apage;		/* an MLGF page */
#ifndef MBR_MLGF_BUFFER
	mlgf_Page	apageBuf;	/* buffer for an MLGF page */

	apage = &apageBuf;
#endif  /* MBR_MLGF_BUFFER */


	/* Read the current page into the buffer. */
#ifdef MBR_MLGF_BUFFER
	e = BfM_GetPage(OPENFILE_DSMSEGID(mlgfd), &pageID, (char**)&apage);
	if (e < 0) ERR(e);
#else
	e = mlgf_ReadPage(mlgfd, pageID, (char*)apage);
	if (e < 0) ERR(e);
#endif  /* MBR_MLGF_BUFFER */

	if (apage->any.hdr.type & MLGF_LEAFPAGE) {
		/* dump leaf patge */
		if(!recordOnly)
		{
			fprintf(fd, "\n <depth : leaf> Leaf Page : %d", pageID.pageNo);
		}
		e = mlgf_DumpLeafPage(mlgfd, &pageID, recordOnly, fd);
#ifdef MBR_MLGF_BUFFER
		if (e < eNOERROR) ERRB1(e, mlgfd, &pageID);
#else
		if (e < eNOERROR) ERR(e);
#endif  /* MBR_MLGF_BUFFER */

	}
	else {
		/* dump directory page */
		if(!recordOnly)
		{
			fprintf(fd, "\n <depth : %d> Directory Page : %d", apage->directory.hdr.height, pageID.pageNo);
			mlgf_DumpDirectoryPage( mlgfd, &pageID, fd);
		}

		for( i=0; i< apage->directory.hdr.nEntries; i++) {
			/* dump next entry */
			entryPtr = MLGF_ITH_DIRENTRY(&apage->directory, i, MLGF_DIRENTRY_LENGTH(OPENFILE_CONTROL(mlgfd).numOfKeys));
			nextPageID.volNo = pageID.volNo;
			nextPageID.pageNo = entryPtr->spid;

			/* call recursion */
			e = mlgf_DumpRecursive( mlgfd, nextPageID, recordOnly, fd);
#ifdef MBR_MLGF_BUFFER
			if (e < eNOERROR) ERRB1(e, mlgfd, &pageID);
#else
			if (e < eNOERROR) ERR( e);
#endif  /* MBR_MLGF_BUFFER */
		}
	}

#ifdef MBR_MLGF_BUFFER  
	e = BfM_FreePage(OPENFILE_DSMSEGID(mlgfd), &pageID);
	if (e < 0) ERR(e);  
#endif  /* MBR_MLGF_BUFFER */

	return(eNOERROR); 
}
