/******************************************************************************/
/*                                                                            */
/*    The Multilevel Grid File (MLGF)                                         */
/*    Version 4.0                                                             */
/*                                                                            */
/*    Developed by Professor Kyu-Young Whang et al. (1994-2016)               */
/*                                                                            */
/*    Advanced Information Technology Research Center (AITrc)                 */
/*    Korea Advanced Institute of Science and Technology (KAIST)              */
/*                                                                            */
/*    e-mail: kywhang@gmail.com                                               */
/*                                                                            */
/*    Bibliography:                                                           */
/*    [1] Whang, K. and Krishnamurthy, R., "The Multilevel Grid File - A      */
/*        Dynamic Hierarchical Multidimensional File Structure," In Proc. 2nd */
/*        Int'l Conf. on Database Systems for Advanced Applications, pp.      */
/*        449-459, 1991.                                                      */
/*    [2] Whang, K., Kim, S., and Wiederhold, G., "Dynamic Maintenance of     */
/*        Data Distribution for Selectivity Estimation," The VLDB Journal,    */
/*        Vol. 3, No. 1, pp. 29-51, 1994.                                     */
/*    [3] Lee, J., Lee, Y., and Whang, K., "A Region Splitting Strategy for   */
/*        Physical Database Design of Multidimensional File Organizations,"   */
/*        In Proc. 23rd Int'l Conf. on Very Large Data Bases, pp. 416-425,    */
/*        1997.                                                               */
/*    [4] Song, J., Whang, K., Lee, Y., Lee, M., and Kim, S., "Spatial Join   */
/*        Processing Using Corner Transformation, IEEE Transactions on        */
/*        Knowledge and Data Engineering (TKDE), Vol. 11, No. 4, pp. 688-695, */
/*        1999.                                                               */
/*    [5] Song, J. et al., "The Clustering Property of Corner Transformation  */
/*        for Spatial Database Applications," Information and Software        */
/*        Technology, Vol. 44, No. 7, pp. 419-429, 2002.                      */
/*    [6] Lee, M., Whang, K., Han, W., and Song, I., "Transform-Space View:   */
/*        Performing Spatial Join in the Transform Space Using Original-Space */
/*        Indexes," IEEE Transactions on Knowledge and Data Engineering       */
/*        (TKDE), Vol 18, No. 2, pp. 245-260, 2006.                           */
/*    [7] Dai, H., Whang, K., and Su, H., "Locality of Corner Transformation  */
/*        for Multidimensional Spatial Access Methods," Electronic Notes in   */
/*        Theoretical Computer Science, Vol. 212, pp. 133-148, 2008.          */
/*                                                                            */
/******************************************************************************/

/*
 * Module: MLGF_OpenIndex.c
 *
 * Description:
 *  Open the MLGF index.
 *
 * Exports:
 *  Four MLGF_OpenIndex(One*, Two*)
 *
 * Returns:
 *  Error code
 *    eSYSERR
 *    eBADPARAMETER
 *    eMEMORYALLOCERR
 */


#include "common.h"
#include "DsM.h"


Four MLGF_OpenIndex(
	One		*path,			/* IN full name of the file */
	Two		*mlgfd)			/* OUT mlgf file descriptor */
{
	int				fd;				/* file descriptor */
	off_t			offset;			/* offset */
	PageID			*pageBuffer;	/* page id */
	Four			freePages;		/* number of free pages */
	Four			i;				/* index */
	FreePageList	*freePagePtr;	/* pointer to free page list */
	FreePageList	*tmp;
	struct stat		statBuf;
	ino_t			ino;

	/* Search the open file table for the given file. */
	if(stat(path, &statBuf) < 0) ERR(eSYSERR);
	if(!S_ISREG(statBuf.st_mode) || S_ISDIR(statBuf.st_mode)) ERR(eBADPARAMETER);;

	/* save the inode number */
	ino = statBuf.st_ino;

	for(i = 0; i < maxEntriesOfOpenFileTable; i++) {
		if(ino == OPENFILE_INO(i)) {
			*mlgfd = i;
			OPENFILE_COUNT(i) ++;
			return(eNOERROR);
		}
	}

	*mlgfd = mlgf_GetOpenFileTableEntry();
	if(*mlgfd < 0) ERR(*mlgfd);

	/* open file */
    if((fd = open(path, O_RDWR)) <0) {
		perror(" open (in MLGFopen)");
		ERR(eSYSERR);
	}

	/* go to position where control information was saved */
	if(lseek(fd, 0 - (sizeof(Four)), L_END) <0) {
		perror(" lseek (in MLGFopen)");
		close(fd);
		ERR(eSYSERR);
	}

	/* read the number of free pages */
	if(read(fd, (char*)&freePages, sizeof(Four)) != sizeof(Four)) {
		perror(" read (in MLGFopen)");
		close(fd);
		ERR(eSYSERR);
	}

	/* rewind file */
														//  PAGECOUNT		 CONTROL
	offset = sizeof(Four) + sizeof(PageID)*freePages + sizeof(PageID) + sizeof(Control);
	if(lseek(fd, -offset, L_END) <0) {
		perror(" lseek (in MLGFopen)");
		close(fd);
		ERR(eSYSERR);
	}

	/* initialize control information */
	if(read(fd, (char*)&OPENFILE_CONTROL(*mlgfd), sizeof(Control)) != sizeof(Control)) {
		perror(" read (in MLGFopen)");
		close(fd);
		ERR(eSYSERR);
	}

	/* read pageCount */
	if(read(fd, (char*)&OPENFILE_PAGECOUNT(*mlgfd), sizeof(PageID)) != sizeof(PageID)) {
		perror(" read (in MLGFopen)");
		close(fd);
		ERR(eSYSERR);
	}

	/* read free page list information */
	if(freePages == 0) {
	    /* make empty free page list */
	    OPENFILE_FREEPAGELIST(*mlgfd).next = NULL;

	} else {
	    pageBuffer = (PageID *)malloc(sizeof(PageID) * freePages);
	    if(pageBuffer == NULL) {
			close(fd);
			ERR(eMEMORYALLOCERR);
	    }
	    if(read(fd, (char*)pageBuffer, sizeof(PageID)*freePages) != (int)sizeof(PageID)*freePages) {
			perror(" read (in MLGFopen)");
			free(pageBuffer);
			close(fd);
			ERR(eSYSERR);
	    }

	    /* make free page list */
	    freePagePtr = &OPENFILE_FREEPAGELIST(*mlgfd);
	    for(i = 0; i< freePages; i++) {
			/* allocate memory */
			freePagePtr->next = (FreePageList *)malloc(sizeof(FreePageList));
			if(freePagePtr->next == NULL) {
				/* free the partially allocated spaces */
				freePagePtr = OPENFILE_FREEPAGELIST(*mlgfd).next;
				while (freePagePtr != NULL) {
					tmp = freePagePtr;
					freePagePtr = freePagePtr->next;
					free(tmp);
				}
				free(pageBuffer);
				close(fd);
				ERR(eMEMORYALLOCERR);
			}
			freePagePtr = freePagePtr->next;
			freePagePtr->pid = pageBuffer[i];
	    }
	    freePagePtr->next = NULL;

	    free(pageBuffer);
	}

#ifdef MBR_MLGF_BUFFER
	close(fd);
	OPENFILE_DSMSEGID(*mlgfd) = dsm_segid_new();
	if(dsm_segid_insert(OPENFILE_DSMSEGID(*mlgfd), path) == eNOERROR &&
         DsM_OpenSegment(OPENFILE_DSMSEGID(*mlgfd), INOUT) == eNOERROR)
	{
		fd = dsm_segid_get(OPENFILE_DSMSEGID(*mlgfd));	/* modify file descriptor */
	}
	else
	{
		ERR(eSYSERR);
	}
#endif  /* MBR_MLGF_BUFFER */

	/* save file descriptor */
	OPENFILE_OS_FD(*mlgfd) = fd;
	OPENFILE_COUNT(*mlgfd) ++;
	OPENFILE_INO(*mlgfd) = ino;

	

	return(eNOERROR);
} /* MLGF_OpenIndex() */
