/******************************************************************************/
/*                                                                            */
/*    The Multilevel Grid File (MLGF)                                         */
/*    Version 4.0                                                             */
/*                                                                            */
/*    Developed by Professor Kyu-Young Whang et al. (1994-2016)               */
/*                                                                            */
/*    Advanced Information Technology Research Center (AITrc)                 */
/*    Korea Advanced Institute of Science and Technology (KAIST)              */
/*                                                                            */
/*    e-mail: kywhang@gmail.com                                               */
/*                                                                            */
/*    Bibliography:                                                           */
/*    [1] Whang, K. and Krishnamurthy, R., "The Multilevel Grid File - A      */
/*        Dynamic Hierarchical Multidimensional File Structure," In Proc. 2nd */
/*        Int'l Conf. on Database Systems for Advanced Applications, pp.      */
/*        449-459, 1991.                                                      */
/*    [2] Whang, K., Kim, S., and Wiederhold, G., "Dynamic Maintenance of     */
/*        Data Distribution for Selectivity Estimation," The VLDB Journal,    */
/*        Vol. 3, No. 1, pp. 29-51, 1994.                                     */
/*    [3] Lee, J., Lee, Y., and Whang, K., "A Region Splitting Strategy for   */
/*        Physical Database Design of Multidimensional File Organizations,"   */
/*        In Proc. 23rd Int'l Conf. on Very Large Data Bases, pp. 416-425,    */
/*        1997.                                                               */
/*    [4] Song, J., Whang, K., Lee, Y., Lee, M., and Kim, S., "Spatial Join   */
/*        Processing Using Corner Transformation, IEEE Transactions on        */
/*        Knowledge and Data Engineering (TKDE), Vol. 11, No. 4, pp. 688-695, */
/*        1999.                                                               */
/*    [5] Song, J. et al., "The Clustering Property of Corner Transformation  */
/*        for Spatial Database Applications," Information and Software        */
/*        Technology, Vol. 44, No. 7, pp. 419-429, 2002.                      */
/*    [6] Lee, M., Whang, K., Han, W., and Song, I., "Transform-Space View:   */
/*        Performing Spatial Join in the Transform Space Using Original-Space */
/*        Indexes," IEEE Transactions on Knowledge and Data Engineering       */
/*        (TKDE), Vol 18, No. 2, pp. 245-260, 2006.                           */
/*    [7] Dai, H., Whang, K., and Su, H., "Locality of Corner Transformation  */
/*        for Multidimensional Spatial Access Methods," Electronic Notes in   */
/*        Theoretical Computer Science, Vol. 212, pp. 133-148, 2008.          */
/*                                                                            */
/******************************************************************************/

/*
 * Module: mlgf_BuddyTest.c
 *
 * Description:
 *  Test if the given two entries are buddy.
 *
 * Exports:
 *  Boolean mlgf_BuddyTest(One, mlgf_DirectoryEntry*, mlgf_DirectoryEntry*, One*)
 *
 * Returns:
 *  TRUE if two entries are buddy entries
 *  FALSE otherwise
 */


#include "common.h"


Boolean mlgf_BuddyTest(
    One                 nKeys,          /* IN number of keys of MLGF index */
    mlgf_DirectoryEntry *entry_x,       /* IN one entry */
    mlgf_DirectoryEntry *entry_y,       /* IN other entry */
    One                 *buddyKey)      /* OUT buddy key no */
{
    One                 i;              /* index */
    One                 buddyMatch;     /* number of buddy matches */
    MLGF_HashValue      *hx;            /* pointer to array of hash values of entry_x */
    MLGF_HashValue      *hy;            /* pointer to array of hash values of entry_y */
    MLGF_HashValue      xor;            /* XOR value of two hash values */


    /* initialization */
    buddyMatch = 0;

    /* hx/hy points to array of hash values of entry_x/entry_y. */
    hx = MLGF_DIRENTRY_HASHVALUEPTR(entry_x, nKeys);
    hy = MLGF_DIRENTRY_HASHVALUEPTR(entry_y, nKeys);

    /* for all the keys */
    for (i = 0; i < nKeys; i++) {
	/* do buddy prefix test */

	if (entry_x->nValidBits[i] != entry_y->nValidBits[i]) return(FALSE);

	xor = (*hx ^ *hy) >> (MLGF_MAXNUM_VALIDBITS - entry_x->nValidBits[i]);

	if (xor == 1) {
	    if (buddyMatch++) return(FALSE);
	    *buddyKey = i;
	} else if (xor != 0)
	    return(FALSE);

        hx++; hy++;
    }

    if (buddyMatch == 0) return(FALSE);

    /* Buddys are regions which were splitted from the same region. */
    for (i = 0; i < nKeys-1; i++)
	if (entry_x->nValidBits[i] != entry_x->nValidBits[i+1]) break;

    if (*buddyKey != i) return(FALSE);

    return TRUE;

} /* mlgf_BuddyTest() */
