/******************************************************************************/
/*                                                                            */
/*    The Multilevel Grid File (MLGF)                                         */
/*    Version 4.0                                                             */
/*                                                                            */
/*    Developed by Professor Kyu-Young Whang et al. (1994-2016)               */
/*                                                                            */
/*    Advanced Information Technology Research Center (AITrc)                 */
/*    Korea Advanced Institute of Science and Technology (KAIST)              */
/*                                                                            */
/*    e-mail: kywhang@gmail.com                                               */
/*                                                                            */
/*    Bibliography:                                                           */
/*    [1] Whang, K. and Krishnamurthy, R., "The Multilevel Grid File - A      */
/*        Dynamic Hierarchical Multidimensional File Structure," In Proc. 2nd */
/*        Int'l Conf. on Database Systems for Advanced Applications, pp.      */
/*        449-459, 1991.                                                      */
/*    [2] Whang, K., Kim, S., and Wiederhold, G., "Dynamic Maintenance of     */
/*        Data Distribution for Selectivity Estimation," The VLDB Journal,    */
/*        Vol. 3, No. 1, pp. 29-51, 1994.                                     */
/*    [3] Lee, J., Lee, Y., and Whang, K., "A Region Splitting Strategy for   */
/*        Physical Database Design of Multidimensional File Organizations,"   */
/*        In Proc. 23rd Int'l Conf. on Very Large Data Bases, pp. 416-425,    */
/*        1997.                                                               */
/*    [4] Song, J., Whang, K., Lee, Y., Lee, M., and Kim, S., "Spatial Join   */
/*        Processing Using Corner Transformation, IEEE Transactions on        */
/*        Knowledge and Data Engineering (TKDE), Vol. 11, No. 4, pp. 688-695, */
/*        1999.                                                               */
/*    [5] Song, J. et al., "The Clustering Property of Corner Transformation  */
/*        for Spatial Database Applications," Information and Software        */
/*        Technology, Vol. 44, No. 7, pp. 419-429, 2002.                      */
/*    [6] Lee, M., Whang, K., Han, W., and Song, I., "Transform-Space View:   */
/*        Performing Spatial Join in the Transform Space Using Original-Space */
/*        Indexes," IEEE Transactions on Knowledge and Data Engineering       */
/*        (TKDE), Vol 18, No. 2, pp. 245-260, 2006.                           */
/*    [7] Dai, H., Whang, K., and Su, H., "Locality of Corner Transformation  */
/*        for Multidimensional Spatial Access Methods," Electronic Notes in   */
/*        Theoretical Computer Science, Vol. 212, pp. 133-148, 2008.          */
/*                                                                            */
/******************************************************************************/

/*
 * Module: mlgf_ChangeLeafEntrySize.c
 *
 * Description:
 *  Change a leaf entry size.
 *
 * Exports:
 *  Four mlgf_ChangeLeafEntrySize(mlgf_LeafPage*, One, Two, Two, Two)
 *
 * Returns:
 *  error code
 *
 * Assumption:
 *  We assume that the page can be accomdate the new length.
 */


#include "common.h"


Four mlgf_ChangeLeafEntrySize(
    mlgf_LeafPage       *apage,         /* INOUT page containing the object */
    One                 nKeys,          /* IN # of keys */
    Two                 entryNo,        /* IN entry number */
    Two                 oldLength,      /* IN old length */
    Two                 newLength,      /* IN new length */
    Two			mlgfd)					/* IN MLGF file descriptor */
{
    mlgf_LeafEntry      *entry;         /* point to the object in page */
    Two                 offset;         /* offset in data area of the object */
    Two                 deltaLength;    /* difference between old length and new length */


    deltaLength = newLength - oldLength;

    /* Points to the object. */
    offset = apage->slot[-entryNo];
    entry = (mlgf_LeafEntry *)&(apage->data[offset]);

    if (deltaLength > 0) {	/* grow the size */

	if (apage->hdr.free == offset+oldLength &&
	    deltaLength <= MLGF_LP_CFREE(apage)) {
	    /* This entry is the last entry in the page and
	     * the added bytes can be appended without movement. */

	    apage->hdr.free += deltaLength;

	} else if (newLength <= MLGF_LP_CFREE(apage)) {
	    /* The contiguous free space can accomadate full entry */

	    /* copy the original entry to the last data space */
	    memcpy(&(apage->data[apage->hdr.free]), entry, oldLength);

	    apage->slot[-entryNo] = apage->hdr.free;
	    apage->hdr.free += newLength;
	    apage->hdr.unused += oldLength;

	} else {
	    /* Complex Case: Compact the data page and insert it */

	    (void) mlgf_CompactLeafPage(apage, nKeys,mlgfd, entryNo);

	    apage->hdr.free += deltaLength;
	}

    } else { /* shrink the size */

	if (apage->hdr.free == offset+oldLength) {
	    /* this is the last entry in the page */
	    /* the space can be reused in the contiguous space */
	    /* apage->hdr.free -= -deltaLength */
	    apage->hdr.free += deltaLength;
	} else {
	    /* free the space: apage->hdr.unused += -deltaLength */
	    apage->hdr.unused -= deltaLength;
	}
    }

    return(eNOERROR);

} /* mlgf_ChangeLeafEntrySize() */
