/******************************************************************************/
/*                                                                            */
/*    The Multilevel Grid File (MLGF)                                         */
/*    Version 4.0                                                             */
/*                                                                            */
/*    Developed by Professor Kyu-Young Whang et al. (1994-2016)               */
/*                                                                            */
/*    Advanced Information Technology Research Center (AITrc)                 */
/*    Korea Advanced Institute of Science and Technology (KAIST)              */
/*                                                                            */
/*    e-mail: kywhang@gmail.com                                               */
/*                                                                            */
/*    Bibliography:                                                           */
/*    [1] Whang, K. and Krishnamurthy, R., "The Multilevel Grid File - A      */
/*        Dynamic Hierarchical Multidimensional File Structure," In Proc. 2nd */
/*        Int'l Conf. on Database Systems for Advanced Applications, pp.      */
/*        449-459, 1991.                                                      */
/*    [2] Whang, K., Kim, S., and Wiederhold, G., "Dynamic Maintenance of     */
/*        Data Distribution for Selectivity Estimation," The VLDB Journal,    */
/*        Vol. 3, No. 1, pp. 29-51, 1994.                                     */
/*    [3] Lee, J., Lee, Y., and Whang, K., "A Region Splitting Strategy for   */
/*        Physical Database Design of Multidimensional File Organizations,"   */
/*        In Proc. 23rd Int'l Conf. on Very Large Data Bases, pp. 416-425,    */
/*        1997.                                                               */
/*    [4] Song, J., Whang, K., Lee, Y., Lee, M., and Kim, S., "Spatial Join   */
/*        Processing Using Corner Transformation, IEEE Transactions on        */
/*        Knowledge and Data Engineering (TKDE), Vol. 11, No. 4, pp. 688-695, */
/*        1999.                                                               */
/*    [5] Song, J. et al., "The Clustering Property of Corner Transformation  */
/*        for Spatial Database Applications," Information and Software        */
/*        Technology, Vol. 44, No. 7, pp. 419-429, 2002.                      */
/*    [6] Lee, M., Whang, K., Han, W., and Song, I., "Transform-Space View:   */
/*        Performing Spatial Join in the Transform Space Using Original-Space */
/*        Indexes," IEEE Transactions on Knowledge and Data Engineering       */
/*        (TKDE), Vol 18, No. 2, pp. 245-260, 2006.                           */
/*    [7] Dai, H., Whang, K., and Su, H., "Locality of Corner Transformation  */
/*        for Multidimensional Spatial Access Methods," Electronic Notes in   */
/*        Theoretical Computer Science, Vol. 212, pp. 133-148, 2008.          */
/*                                                                            */
/******************************************************************************/

/*
 * Module: mlgf_DeleteObjectFromLeaf.c
 *
 * Description:
 *  Delete the given object from a leaf page.
 */


#include "common.h"


Four mlgf_DeleteObjectFromLeaf(
	Two							mlgfd,					/* IN MLGF file descriptor */
    PageID                      *root,                  /* IN root of subtree */ 
    mlgf_LeafPage               *apage,                 /* INOUT leaf page */
    MLGF_KeyDesc                *kdesc,                 /* IN key descriptor of MLGF index */
    MLGF_HashValue              keys[],                 /* IN hash values of the delete object */
	Object						*obj,					/* IN object to delete */
    mlgf_DirectoryEntry         *entryToLeaf,           /* INOUT entry for the leaf node in parent node */
    mlgf_DeleteStatus_T         *status)                /* INOUT program status */
{
    Four                        e;                      /* error code */
    Two                         i;                      /* index variable */
    Four                        of;                     /* # of objects in overflow apge when underflow */  
	Four						objArrayElemOffset;		/* offset on object array */
    Two                         objArrayElemNo;         /* index on object array */
    Two                         entryNo;                /* slot No. of entry with the given keys */
    Two                         entryOldLen;            /* length of an entry before object deletion */
    Two                         entryNewLen;            /* length of an entry after object deletion */
    Two                         objectItemLen;          /* length of an object item in object array */
    char                        *objectItemPtr;         /* points to object item which will be deleted */
    PageID                      ovPid;                  /* PageID of the first overflow page of an entry */
    Boolean                     found;                  /* TRUE if we find something to want */
    mlgf_LeafEntry              *entry;                 /* points to leaf entry */
    mlgf_OverflowPage           *opage;                 /* points to an overflow page */
    mlgf_MortonValue            keyMortonVal;           /* mroton value for the key */
    One                         nValidBits[MLGF_MAXNUM_KEYS]; /* used for getting morton value */
    MLGF_HashValue              *dirEntryHashValues;    /* points to arrary of hash values in an entry */
    One                         k;
    Two				nObjects;
    Two				totalObjLen;
#ifndef MBR_MLGF_BUFFER
    mlgf_OverflowPage           opageBuf;				/* buffer for an overflow page */

	opage = &opageBuf;
#endif  /* MBR_MLGF_BUFFER */


    /* find object to delete */

    /* Get the morton value of the keys. */
    for (k = 0; k < kdesc->nKeys; k++)
        nValidBits[k] = MLGF_MAXNUM_VALIDBITS;

    mlgf_GetMortonValue(keys, nValidBits, &keyMortonVal, kdesc->nKeys);

    /* Search the leaf page for the entry whose keys are equal to the given keys. */
    found = mlgf_SearchLeafPageInMortonOrder(apage, kdesc, &keyMortonVal, &entryNo);

	if (!found) {
		status->flags.notFound = 1;
		return(eNOERROR);
	}

    /* Get the length of an object item in object array. */
     objectItemLen = mlgf_leafEntryObjectItemLen(mlgfd,obj);

    /* 'entry' points to the leaf entry which contains the deleted object. */
    entry = MLGF_ITH_LEAFENTRY(apage, entryNo);

    if (mlgf_leafGetNObjects(mlgfd,entry) > 0) { /* normal entry */

	found = mlgf_SearchObjectArray(MLGF_LEAFENTRY_FIRST_OBJECT(OPENFILE_CONTROL(mlgfd).useAdditionalFunc,kdesc->nKeys, entry), obj, mlgf_leafGetNObjects(mlgfd,entry), &objArrayElemNo, &objArrayElemOffset,mlgfd);

	if (!found) {
            status->flags.notFound = 1;
            return(eNOERROR);
        }

	/* 'objectItemPtr' points to the deleted object. */
	objectItemPtr = (char*)(MLGF_LEAFENTRY_FIRST_OBJECT(OPENFILE_CONTROL(mlgfd).useAdditionalFunc,kdesc->nKeys, entry) + objArrayElemOffset);

	 if (mlgf_leafGetNObjects(mlgfd,entry) == 1) {

	    memmove(&apage->slot[-(apage->hdr.nEntries-2)], &apage->slot[-(apage->hdr.nEntries-1)],
		    (apage->hdr.nEntries-entryNo-1)*sizeof(Two));
	    apage->hdr.nEntries--;

 	    apage->hdr.unused += MLGF_LEAFENTRY_LENGTH(OPENFILE_CONTROL(mlgfd).useAdditionalFunc,kdesc->nKeys, mlgf_leafGetTotalObjLen(mlgfd,entry), mlgf_leafGetNObjects(mlgfd,entry));

	} else {

	    memmove(objectItemPtr, objectItemPtr + objectItemLen,mlgf_leafGetTotalObjLen(mlgfd,entry) - (objArrayElemOffset + objectItemLen));
				
		totalObjLen = mlgf_leafGetTotalObjLen(mlgfd,entry);
		mlgf_leafSetTotalObjLen(mlgfd,entry,totalObjLen - objectItemLen);

	    nObjects = mlgf_leafGetNObjects(mlgfd,entry);
	    mlgf_leafSetNObjects(mlgfd,entry,nObjects-1);
	    apage->hdr.unused += objectItemLen;
	}

#ifdef MBR_MLGF_BUFFER
        /* Set the dirty flag. */
        e = BfM_SetDirty(OPENFILE_DSMSEGID(mlgfd), &apage->hdr.pid);
        if (e < 0) ERR(e);
#else
		e = mlgf_WritePage(mlgfd, apage->hdr.pid, (char*)apage);
		if (e < 0) ERR(e);
#endif  /* MBR_MLGF_BUFFER */

    } else {		   /* overflow page */
	MAKE_PAGEID(ovPid, apage->hdr.pid.volNo, MLGF_LEAFENTRY_FIRST_OVERFLOW(kdesc->nKeys, entry));
	e = mlgf_DeleteOverflow(mlgfd, root, &ovPid, kdesc, keys, obj, &of); 
	if (e == eNOTFOUND) {
		status->flags.notFound = 1;
		return(eNOERROR);
	}
	else if (e < 0) ERR(e);

        entryOldLen = MLGF_LEAFENTRY_LENGTH(OPENFILE_CONTROL(mlgfd).useAdditionalFunc,kdesc->nKeys, mlgf_leafGetTotalObjLen(mlgfd,entry), mlgf_leafGetNObjects(mlgfd,entry));
        entryNewLen = MLGF_NEW_LEAFENTRY_LENGTH(OPENFILE_CONTROL(mlgfd).useAdditionalFunc,kdesc->nKeys, of);

        if ((entryNewLen < MLGF_OP_OUT_THRESHOLD) && (entryNewLen - entryOldLen <= (Two)MLGF_LP_FREE(apage))) {
#ifdef MBR_MLGF_BUFFER
            /*@ get the page */
            e = BfM_GetPage(OPENFILE_DSMSEGID(mlgfd), &ovPid, (char**)&opage);
            if (e < eNOERROR) ERR(e);
#else
			e = mlgf_ReadPage(mlgfd, ovPid, (char*)opage);
			if (e < eNOERROR) ERR(e);
#endif  /* MBR_MLGF_BUFFER */

            if (of == 0) { /* Delete the leaf entry. */
                /* Move slots to delete the slot */
                for (i = entryNo+1; i < apage->hdr.nEntries; i++)
                    apage->slot[-(i-1)] = apage->slot[-i];

                /* free the space and decrement the # of entries by 1 */
                apage->hdr.unused += entryOldLen;
                apage->hdr.nEntries --;

            } else {    /* of > 0 */
                mlgf_ChangeLeafEntrySize(apage, kdesc->nKeys, entryNo, entryOldLen, entryNewLen,mlgfd);

                /* entry was moved on the above call */
                entry = MLGF_ITH_LEAFENTRY(apage, entryNo);

                /* the original leaf entry is on free space boundary */
				if ( OPENFILE_CONTROL(mlgfd).useAdditionalFunc)
					memcpy((char*)((entry)->keys + kdesc->nKeys+1), (char*)opage->data, of);
				else
					memcpy((char*)((entry)->keys + kdesc->nKeys), (char*)opage->data, of);

			mlgf_leafSetTotalObjLen(mlgfd,entry,of);
		mlgf_leafSetNObjects(mlgfd,entry,opage->hdr.nObjects);
            }

#ifdef MBR_MLGF_BUFFER
            e = BfM_FreePage(OPENFILE_DSMSEGID(mlgfd), &ovPid);
            if (e < eNOERROR) ERR(e);
#else
#endif  /* MBR_MLGF_BUFFER */

			e = mlgf_FreePage(mlgfd, ovPid);
			if (e < 0) ERR(e);

#ifdef MBR_MLGF_BUFFER
            /* Set the dirty flag. */
            e = BfM_SetDirty(OPENFILE_DSMSEGID(mlgfd), &apage->hdr.pid);
            if (e < 0) ERR(e);
#else
			e = mlgf_WritePage(mlgfd, apage->hdr.pid, (char*)apage);
			if (e < 0) ERR(e);
#endif  /* MBR_MLGF_BUFFER */
        }
    }

    /*
    ** Update the extreme hash values of the parent entry.
    */
    dirEntryHashValues = MLGF_DIRENTRY_HASHVALUEPTR(entryToLeaf, kdesc->nKeys);
    for (k = 0; k < kdesc->nKeys; k++) {

	if ((MLGF_KEYDESC_IS_MINTYPE(*kdesc, k) && (dirEntryHashValues[k] < keys[k])) ||
	    (MLGF_KEYDESC_IS_MAXTYPE(*kdesc, k) && (dirEntryHashValues[k] > keys[k])))
	    continue;

        status->flags.mbrUpdated = 1;
	if (MLGF_KEYDESC_IS_MINTYPE(*kdesc, k))
            dirEntryHashValues[k] |= MLGF_HASHVALUE_SET_EXCEPT_UPPER_N_BITS(entryToLeaf->nValidBits[k]);
        else
            dirEntryHashValues[k] = MLGF_HASHVALUE_MASK_UPPER_N_BITS(dirEntryHashValues[k], entryToLeaf->nValidBits[k]);

	for (i = 0; i < apage->hdr.nEntries; i++) {
	    entry = MLGF_ITH_LEAFENTRY(apage, i);

	    if ((MLGF_KEYDESC_IS_MINTYPE(*kdesc, k) && (dirEntryHashValues[k] > entry->keys[k])) ||
		(MLGF_KEYDESC_IS_MAXTYPE(*kdesc, k) && (dirEntryHashValues[k] < entry->keys[k]))) {
		dirEntryHashValues[k] = entry->keys[k];
	    }
	}
    }

    /* Set the theta value. */
    if (MLGF_LP_THETA(apage) != entryToLeaf->theta) {
        entryToLeaf->theta = MLGF_LP_THETA(apage);
        status->flags.thetaUpdated = 1;
    }

    if (apage->hdr.nEntries == 0)
        status->flags.emptyPage = 1;
    else if (MLGF_LP_THETA(apage) < MLGF_LP_THRESHOLD)
	{
        status->flags.underflow = 1;
	}

    return(eNOERROR);

} /* mlgf_DeleteObjectFromLeaf() */
